@@ -1,8 +1,20 @@
-#![allow(clippy::not_unsafe_ptr_arg_deref)]
+//! Headers and top line of an http object
+//!
+//! Depending on the VCL subroutine, the `Ctx` will give access to various [`HTTP`] object which
+//! expose the request line (`req`, `req_top` and `bereq`), response line (`resp`, `beresp`) and
+//! headers of the objects Varnish is manipulating.
+//!
+//! `HTTP` implements `IntoIterator` that will expose the headers only (not the `method`, `status`,
+//! etc.)
+//!
+//! **Note:** at this stage, headers are asumed to be utf8, and you will get a panic if it's not
+//! the case. Future work needs to sanitize the headers to make this safer to use. It is tracked in
+//! this [issue](https://github.com/gquintard/varnish-rs/issues/4).
 
+#![allow(clippy::not_unsafe_ptr_arg_deref)]
+use std::os::raw::c_uint;
 use std::slice::{from_raw_parts, from_raw_parts_mut};
 use std::str::from_utf8;
-use std::os::raw::c_uint;
 
 use crate::vcl::ws::WS;
 
@@ -63,8 +75,11 @@ impl<'a> HTTP<'a> {
         let res = self.change_header(idx, name, value);
         if res.is_ok() {
             unsafe {
-                varnish_sys::VSLbt(self.raw.vsl, self.raw.logtag as c_uint + HDR_FIRST as c_uint,
-                                   *self.raw.hd.add(idx as usize));
+                varnish_sys::VSLbt(
+                    self.raw.vsl,
+                    self.raw.logtag as c_uint + HDR_FIRST as c_uint,
+                    *self.raw.hd.add(idx as usize),
+                );
             }
         } else {
             self.raw.nhd -= 1;
@@ -82,8 +97,11 @@ impl<'a> HTTP<'a> {
             let (n, _) = header_from_hd(hd).unwrap();
             if name.eq_ignore_ascii_case(n) {
                 unsafe {
-                    varnish_sys::VSLbt(self.raw.vsl, self.raw.logtag + HDR_UNSET as u32 - HDR_METHOD as u32,
-                                       *self.raw.hd.add(HDR_FIRST as usize + idx as usize));
+                    varnish_sys::VSLbt(
+                        self.raw.vsl,
+                        self.raw.logtag + HDR_UNSET as u32 - HDR_METHOD as u32,
+                        *self.raw.hd.add(HDR_FIRST as usize + idx as usize),
+                    );
                 }
                 continue;
             }
